# acb.py: For all your ACB extracting needs

# Copyright (c) 2016, The Holy Constituency of the Summer Triangle.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# This program is based on code from VGMToolbox.
# Copyright (c) 2009
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import io
import math
import struct
import sys
import itertools
import os
import re
from collections import namedtuple as T
from typing import Optional, Union, BinaryIO, Tuple, List, Callable

from .utf import UTFTable, R
from .disarm import DisarmContext

WAVEFORM_ENCODE_TYPE_ADX          = 0
WAVEFORM_ENCODE_TYPE_HCA          = 2
WAVEFORM_ENCODE_TYPE_VAG          = 7
WAVEFORM_ENCODE_TYPE_ATRAC3       = 8
WAVEFORM_ENCODE_TYPE_BCWAV        = 9
WAVEFORM_ENCODE_TYPE_NINTENDO_DSP = 13

wave_type_ftable = {
    WAVEFORM_ENCODE_TYPE_ADX          : ".adx",
    WAVEFORM_ENCODE_TYPE_HCA          : ".hca",
    WAVEFORM_ENCODE_TYPE_VAG          : ".vag",
    WAVEFORM_ENCODE_TYPE_ATRAC3       : ".at3",
    WAVEFORM_ENCODE_TYPE_BCWAV        : ".bcwav",
    WAVEFORM_ENCODE_TYPE_NINTENDO_DSP : ".dsp"}

track_t = T("track_t", ("cue_id", "name", "memory_wav_id", "external_wav_id", "enc_type", "is_stream"))

class TrackList(object):
    def __init__(self, utf):
        cue_handle = io.BytesIO(utf.rows[0]["CueTable"])
        nam_handle = io.BytesIO(utf.rows[0]["CueNameTable"])
        wav_handle = io.BytesIO(utf.rows[0]["WaveformTable"])
        syn_handle = io.BytesIO(utf.rows[0]["SynthTable"])

        cues = UTFTable(cue_handle)
        nams = UTFTable(nam_handle)
        wavs = UTFTable(wav_handle)
        syns = UTFTable(syn_handle)

        self.tracks: List[track_t] = []

        name_map = {}
        for row in nams.rows:
            name_map[row["CueIndex"]] = row["CueName"]

        for ind, row in enumerate(cues.rows):
            if row["ReferenceType"] not in {3, 8}:
                raise RuntimeError("ReferenceType {0} not implemented.".format(row["ReferenceType"]))

            r_data = syns.rows[row["ReferenceIndex"]]["ReferenceItems"]
            a, b = struct.unpack(">HH", r_data)

            wav_id = wavs.rows[b].get("Id")
            if wav_id is None:
                wav_id = wavs.rows[b]["MemoryAwbId"]
            extern_wav_id = wavs.rows[b]["StreamAwbId"]
            enc = wavs.rows[b]["EncodeType"]
            is_stream = wavs.rows[b]["Streaming"]

            self.tracks.append(track_t(row["CueId"], name_map.get(ind, "UNKNOWN"), wav_id,
                extern_wav_id, enc, is_stream))

def align(n):
    def _align(number):
        return (number + n - 1) & ~(n - 1)
    return _align

afs2_file_ent_t = T("afs2_file_ent_t", ("cue_id", "offset", "size"))

class AFSArchive(object):
    def __init__(self, file):
        buf = R(file)

        magic = buf.uint32_t()
        if magic != 0x41465332:
            raise ValueError("bad magic")

        version = buf.bytes(4)
        file_count = buf.le_uint32_t()

        if version[0] >= 0x02:
            self.alignment = buf.le_uint16_t()
            self.mix_key = buf.le_uint16_t()
        else:
            self.alignment = buf.le_uint32_t()
            self.mix_key = None

        #print("afs2:", file_count, "files in ar")
        #print("afs2: aligned to", self.alignment, "bytes")

        self.offset_size = version[1]
        self.offset_mask = int("FF" * self.offset_size, 16)
        #print("afs2: a file offset is", self.offset_size, "bytes")

        self.files = []
        self.create_file_entries(buf, file_count)
        self.src = buf

    def create_file_entries(self, buf, file_count):
        buf.seek(0x10)
        read_cue_ids = struct.Struct("<" + ("H" * file_count))
        if self.offset_size == 2:
            read_raw_offs = struct.Struct("<" + ("H" * (file_count + 1)))
        else:
            read_raw_offs = struct.Struct("<" + ("I" * (file_count + 1)))

        # read all in one go
        cue_ids = buf.struct(read_cue_ids)
        raw_offs = buf.struct(read_raw_offs)
        # apply the mask
        unaligned_offs = tuple(map(lambda x: x & self.offset_mask, raw_offs))
        aligned_offs = tuple(map(align(self.alignment), unaligned_offs))
        offsets_for_length_calculating = unaligned_offs[1:]
        lengths = itertools.starmap(
            lambda my_offset, next_offset: next_offset - my_offset,
            zip(aligned_offs, offsets_for_length_calculating))

        self.files = list(itertools.starmap(afs2_file_ent_t, zip(cue_ids, aligned_offs, lengths)))

    def file_data_for_cue_id(self, cue_id, rw=False):
        for f in self.files:
            if f.cue_id == cue_id:
                if rw:
                    buf = bytearray(f.size)
                    self.src.bytesinto(buf, at=f.offset)
                    return buf
                else:
                    return self.src.bytes(f.size, at=f.offset)
        else:
            raise ValueError("id {0} not found in archive".format(cue_id))

AnyFile = Union[str, os.PathLike, BinaryIO]
Uninitialized = object()

def _get_file_obj(name: AnyFile) -> Tuple[BinaryIO, bool]:
    if isinstance(name, (str, os.PathLike)):
        return open(name, "rb"), True
    else:
        return name, False

class ACBFile(object):
    """ Represents an ACB file.

        You should call the close() method to release the underlying file objects
        when finished, or use the object as a context manager.
        Tracks can be accessed using .track_list together with get_track_data().

        Constructor arguments:
        - acb_file: Path to the file, or an open file object in binary mode.
        - extern_awb: Path to the streaming AWB file, if needed.
        - hca_keys: HCA keys in string format. Can be one of:
            "0xLOWBYTES,0xHIGHBYTES", "0xHIGHBYTESLOWBYTES", or None.
            "0xLOWBYTES,0xHIGHBYTES" is equivalent to invoking hca_decoder
            with arguments "-a LOWBYTES -b HIGHBYTES".
            If None, HCA files will not be decrypted.
    """
    def __init__(self, acb_file: AnyFile, extern_awb: Optional[AnyFile] = None, hca_keys: Optional[str] = None):
        self.acb_handle, self.acb_handle_owned = _get_file_obj(acb_file)
        
        if extern_awb is None:
            self.awb_handle = None
            self.awb_handle_owned = False
        else:
            self.awb_handle, self.awb_handle_owned = _get_file_obj(extern_awb)
                
        utf = UTFTable(self.acb_handle)
        if len(utf.rows[0]["AwbFile"]) > 0:
            self.embedded_awb = AFSArchive(io.BytesIO(utf.rows[0]["AwbFile"]))
        else:
            self.embedded_awb = None # type: ignore

        if self.awb_handle:
            self.external_awb = AFSArchive(self.awb_handle)
        else:
            self.external_awb = None # type: ignore

        self.hca_keys = hca_keys
        self.embedded_disarm: Optional[DisarmContext] = Uninitialized # type: ignore
        self.external_disarm: Optional[DisarmContext] = Uninitialized # type: ignore

        self.track_list = TrackList(utf)
        self.closed = False
    
    def get_embedded_disarm(self) -> Optional[DisarmContext]:
        if self.embedded_disarm is Uninitialized:
            if self.hca_keys:
                self.embedded_disarm = DisarmContext(self.hca_keys, self.embedded_awb.mix_key)
            else:
                self.embedded_disarm = None

        return self.embedded_disarm
    
    def get_external_disarm(self) -> Optional[DisarmContext]:
        if self.external_disarm is Uninitialized:
            if self.hca_keys:
                self.external_disarm = DisarmContext(self.hca_keys, self.external_awb.mix_key)
            else:
                self.external_disarm = None
        return self.external_disarm

    def get_track_data(self, track: track_t, disarm: Optional[bool] = None, unmask: bool = True) -> bytearray:
        """ Gets encoded audio data as a bytearray.

            Arguments:
            - track: The track to get data for, from .track_list.
            - disarm: Whether to decrypt HCA data before returning it.
                The default action is to decrypt if hca_keys were passed when creating
                the ACBFile. You can pass False to skip decryption, or True to force
                decryption. 
                ValueError is raised you force decryption and keys were not passed.
            - unmask: Whether to remove XOR masking from HCA header tags. 
                This only has an effect if decryption is enabled, whether implicitly or
                explicitly.
        """
        if self.closed:
            raise ValueError("ACBFile is closed")

        if track.is_stream:
            buf = self.external_awb.file_data_for_cue_id(track.external_wav_id, rw=True)
            disarmer = self.get_external_disarm()
        else:
            buf = self.embedded_awb.file_data_for_cue_id(track.memory_wav_id, rw=True)
            disarmer = self.get_embedded_disarm()

        if disarm is True and not disarmer:
            raise ValueError(
                "Disarm was explicitly requested, but no keys were provided. "
                "Either remove the disarm= argument from the call to get_track_data, "
                "or provide keys using the hca_keys= argument to ACBFile."
            )
    
        if disarm is None:
            disarm = (disarmer is not None)

        if disarm and disarmer:
            disarmer.disarm(buf, not unmask)

        return buf

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        """ Close any open files held by ACBFile. If you passed file objects
            instead of paths when creating the ACBFile instance, they will
            not be closed.

            Can be called multiple times; subsequent calls will have no effect.
        """
        if self.acb_handle_owned:
            self.acb_handle.close()
            self.acb_handle_owned = False
        if self.awb_handle_owned:
            self.awb_handle.close()
            self.awb_handle_owned = False
        self.closed = True

    def __del__(self):
        self.close()


def find_awb(path):
    return re.sub(r"\.acb$", ".awb", path)

def name_gen_default(track):
    return "{0}{1}".format(track.name, wave_type_ftable.get(track.enc_type, track.enc_type))

def extract_acb(
    acb_file: AnyFile,
    target_dir: str,
    extern_awb: Optional[AnyFile] = None,
    hca_keys: Optional[str] = None,
    name_gen: Callable[[track_t], str] = name_gen_default,
    no_unmask: bool = False
):
    """ Oneshot file extraction API. Dumps all tracks from a file into the
        named output directory.

        Arguments:
        - acb_file: Path to the file, or an open file object in binary mode.
        - target_dir: Path to the destination directory. Must already exist.
        - extern_awb: Path to the streaming AWB file, if needed.
        - hca_keys: Same as ACBFile's hca_keys argument.
        - name_gen: A callable taking the track_t object and returning a 
            destination filename. Should not return absolute paths, as
            they will be prefixed with the target_dir.
        - no_unmask: See ACBFile.get_track_data's unmask argument. For 
            compatibility reasons, the meaning of this flag is reversed;
            i.e. True will result in unmasking being disabled.
    """
    if isinstance(acb_file, str) and extern_awb is None:
        awb_name = find_awb(acb_file)
        if os.path.exists(awb_name):
            extern_awb = awb_name

    with ACBFile(acb_file, extern_awb=extern_awb, hca_keys=hca_keys) as acb:
        for track in acb.track_list.tracks:
            print(track)
            name = name_gen(track)

            with open(os.path.join(target_dir, name), "wb") as out_file:
                out_file.write(acb.get_track_data(track))



import argparse
import os

from acb import extract_acb

def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--disarm-with", help="decrypt HCAs with provided keys")
    parser.add_argument("--awb", help="use file as the external AWB")
    parser.add_argument("--no-unmask", action="store_true", default=False,
        help="don't unmask segment names (requires --disarm-with)")
    parser.add_argument("acb_file", help="input ACB file")
    parser.add_argument("output_dir", help="directory to place output files in")

    args = parser.parse_args()

    os.makedirs(args.output_dir, 0o755, exist_ok=True)
    extract_acb(args.acb_file, args.output_dir, args.awb, args.disarm_with, no_unmask=args.no_unmask)

if __name__ == '__main__':
    main()



import binascii
import struct
import time

try:
    import _acb_speedup
except ImportError:
    _acb_speedup = None

SECTION_SIZES = {
    b"HCA\x00": 8,
    b"fmt\x00": 16,
    b"comp":    16,
    b"dec\x00": 12,
    b"vbr\x00": 8,
    b"ath\x00": 6,
    b"loop":    16,
    b"ciph":    6,
    b"rva\x00": 8,
    b"comm":    5,
    b"pad\x00": 4,
}

CHECKSUM_TABLE = (
    0x0000,0x8005,0x800F,0x000A,0x801B,0x001E,0x0014,0x8011,0x8033,0x0036,0x003C,0x8039,0x0028,0x802D,0x8027,0x0022,
    0x8063,0x0066,0x006C,0x8069,0x0078,0x807D,0x8077,0x0072,0x0050,0x8055,0x805F,0x005A,0x804B,0x004E,0x0044,0x8041,
    0x80C3,0x00C6,0x00CC,0x80C9,0x00D8,0x80DD,0x80D7,0x00D2,0x00F0,0x80F5,0x80FF,0x00FA,0x80EB,0x00EE,0x00E4,0x80E1,
    0x00A0,0x80A5,0x80AF,0x00AA,0x80BB,0x00BE,0x00B4,0x80B1,0x8093,0x0096,0x009C,0x8099,0x0088,0x808D,0x8087,0x0082,
    0x8183,0x0186,0x018C,0x8189,0x0198,0x819D,0x8197,0x0192,0x01B0,0x81B5,0x81BF,0x01BA,0x81AB,0x01AE,0x01A4,0x81A1,
    0x01E0,0x81E5,0x81EF,0x01EA,0x81FB,0x01FE,0x01F4,0x81F1,0x81D3,0x01D6,0x01DC,0x81D9,0x01C8,0x81CD,0x81C7,0x01C2,
    0x0140,0x8145,0x814F,0x014A,0x815B,0x015E,0x0154,0x8151,0x8173,0x0176,0x017C,0x8179,0x0168,0x816D,0x8167,0x0162,
    0x8123,0x0126,0x012C,0x8129,0x0138,0x813D,0x8137,0x0132,0x0110,0x8115,0x811F,0x011A,0x810B,0x010E,0x0104,0x8101,
    0x8303,0x0306,0x030C,0x8309,0x0318,0x831D,0x8317,0x0312,0x0330,0x8335,0x833F,0x033A,0x832B,0x032E,0x0324,0x8321,
    0x0360,0x8365,0x836F,0x036A,0x837B,0x037E,0x0374,0x8371,0x8353,0x0356,0x035C,0x8359,0x0348,0x834D,0x8347,0x0342,
    0x03C0,0x83C5,0x83CF,0x03CA,0x83DB,0x03DE,0x03D4,0x83D1,0x83F3,0x03F6,0x03FC,0x83F9,0x03E8,0x83ED,0x83E7,0x03E2,
    0x83A3,0x03A6,0x03AC,0x83A9,0x03B8,0x83BD,0x83B7,0x03B2,0x0390,0x8395,0x839F,0x039A,0x838B,0x038E,0x0384,0x8381,
    0x0280,0x8285,0x828F,0x028A,0x829B,0x029E,0x0294,0x8291,0x82B3,0x02B6,0x02BC,0x82B9,0x02A8,0x82AD,0x82A7,0x02A2,
    0x82E3,0x02E6,0x02EC,0x82E9,0x02F8,0x82FD,0x82F7,0x02F2,0x02D0,0x82D5,0x82DF,0x02DA,0x82CB,0x02CE,0x02C4,0x82C1,
    0x8243,0x0246,0x024C,0x8249,0x0258,0x825D,0x8257,0x0252,0x0270,0x8275,0x827F,0x027A,0x826B,0x026E,0x0264,0x8261,
    0x0220,0x8225,0x822F,0x022A,0x823B,0x023E,0x0234,0x8231,0x8213,0x0216,0x021C,0x8219,0x0208,0x820D,0x8207,0x0202 )

def checksum(buf):
    sum = 0
    for byte in buf:
        sum = ((sum << 8) ^ CHECKSUM_TABLE[(sum >> 8) ^ byte]) & 0xffff
    return sum

def sub1_rollover(i: int) -> int:
    if i == 0:
        return 0xffffffff
    return i - 1

def small_rng(seed: int) -> bytearray:
    a = ((seed & 1) << 3) | 0b0101
    c = (seed & 0xe) | 1
    seed >>= 4

    table = bytearray(16)
    for i in range(16):
        seed = (seed * a + c) & 0xf
        table[i] = seed
    return table

def mix_header_key(kl, hk):
    k2p = (hk << 16) | (((hk ^ 0xffff) + 2) & 0xffff)
    return (kl * k2p) & 0xffffffffffffffff

class DisarmContext(object):
    KEY_TABLE_1 = None

    @classmethod
    def init_table1(cls) -> bytearray:
        t = bytearray(256)
        t[0x00] = 0
        t[0xff] = 0xff
        seed = 0

        for i in range(1, 255):
            seed = (seed * 13 + 11) & 0xff
            if seed == 0 or seed == 0xff:
                seed = (seed * 13 + 11) & 0xff
            t[i] = seed

        return t

    def __init__(self, keyspec: str, header_key: int = None):
        if "," in keyspec:
            keya_, keyb_ = keyspec.split(",")
            keya = int(keya_, 16)
            keyb = int(keyb_, 16)

            if header_key:
                keylong = mix_header_key(keya | (keyb << 32), header_key)
                keya = keylong & 0xffffffff
                keyb = (keylong >> 32) & 0xffffffff
        else:
            keylong = int(keyspec, 16)
            if header_key:
                keylong = mix_header_key(keylong, header_key)

            keya = keylong & 0xffffffff
            keyb = (keylong >> 32) & 0xffffffff

        self.keya = keya
        self.keyb = keyb
        self.tables()

    def tables(self):
        if self.__class__.KEY_TABLE_1 is None:
            self.__class__.KEY_TABLE_1 = self.init_table1()

        self.key_table_2 = self.init_table2()

    def init_table2(self) -> bytearray:
        keya = self.keya
        keyb = self.keyb

        if keya == 0:
            keyb = sub1_rollover(keyb)
        keya = sub1_rollover(keya)

        stage1 = bytearray(8)
        for i in range(7):
            stage1[i] = keya & 0xff
            keya = ((keya >> 8) | (keyb << 24)) & 0xffffffff
            keyb >>= 8

        s2_ri = (1, 2, 3, 4, 5, 6, 1)
        s2_li = (0, 6, 1, 2, 3, 4, 5)
        stage2 = bytearray(16)

        for i in range(0, 16, 3):
            stage2[i] = stage1[i // 3 + 1]

        for i in range(1, 16, 3):
            j = i // 3 + 1
            stage2[i] = stage1[j] ^ stage1[s2_li[j]]

        for i in range(2, 16, 3):
            j = i // 3 + 2
            stage2[i] = stage1[j] ^ stage1[s2_ri[j]]

        stage3 = bytearray(256)
        high = small_rng(stage1[0])
        for i in range(16):
            base = 16 * i
            low = small_rng(stage2[i])
            hi = high[i] << 4
            for j in range(16):
                stage3[base + j] = hi | low[j]

        key_table_2 = bytearray(256)
        key_table_2[0x00] = 0
        key_table_2[0xff] = 0xff
        srci = 0x11
        for i in range(1, 255):
            while stage3[srci] in (0x0, 0xff):
                srci = (srci + 0x11) & 0xff

            key_table_2[i] = stage3[srci]
            srci = (srci + 0x11) & 0xff

        return key_table_2

    def disarm(self, buf: bytearray, no_unmask: bool=False):
        magic = buf[:4]
        masked = True if magic[0] & 0x80 else False
        header_size = struct.unpack(">H", buf[6:8])[0]

        if not no_unmask:
            self.unmask_header(buf, header_size)
            masked = False

        try:
            comp_seg = buf.index(b"\xe3\xef\xed\xf0" if masked else b"comp", 0, header_size)
        except ValueError:
            try:
                comp_seg = buf.index(b"\xe4\xe5\xe3\x00" if masked else b"dec\x00", 0, header_size)
            except ValueError:
                raise ValueError("cannot find a segment containing the block size")

        try:
            ciph_seg = buf.index(b"\xe3\xe9\xf0\xe8" if masked else b"ciph", 0, header_size)
        except ValueError:
            return

        try:
            fmt_seg = buf.index(b"\xe6\xed\xf4\x00" if masked else b"fmt\x00", 0, header_size)
        except ValueError:
            raise ValueError("cannot find the fmt segment")

        block_cnt = struct.unpack(">I", buf[fmt_seg + 8:fmt_seg + 12])[0]
        block_size = struct.unpack(">H", buf[comp_seg + 4:comp_seg + 6])[0]
        ciph_type = struct.unpack(">H", buf[ciph_seg + 4:ciph_seg + 6])[0]

        if ciph_type == 0:
            return

        if ciph_type == 1:
            self.disarm_actual(buf, header_size, block_cnt, block_size, self.KEY_TABLE_1)
        elif ciph_type == 56:
            self.disarm_actual(buf, header_size, block_cnt, block_size, self.key_table_2)
        else:
            raise ValueError("unknown cipher type")

        buf[ciph_seg + 4:ciph_seg + 6] = b"\x00\x00"

        end = header_size - 2
        if _acb_speedup:
            _acb_speedup.checksum_block_fast(memoryview(buf)[:end])
        else:
            buf[end:end + 2] = checksum(memoryview(buf)[:end]).to_bytes(2, "big")

    def unmask_header(self, buf, header_size):
        base = 0
        while base < header_size:
            tag = bytes(x & 0x7f for x in buf[base:base + 4])
            buf[base:base + 4] = tag

            if tag == b"pad\x00":
                break
            if tag == b"comm":
                base += buf[base + 4]

            base += SECTION_SIZES.get(tag, 4)

    def disarm_actual(self, buf, frompos, blockcnt, blocksize, usetable):
        base = frompos
        stop = frompos + (blocksize * blockcnt)
        while base < stop:
            if _acb_speedup:
                _acb_speedup.disarm_block_fast(memoryview(buf)[base:base + blocksize], usetable)
            else:
                for i in range(base, base + blocksize - 2):
                    buf[i] = usetable[buf[i]]

                end = base + blocksize - 2
                buf[end:end + 2] = checksum(memoryview(buf)[base:end]).to_bytes(2, "big")

            base += blocksize



# CRI UTF table parser

# Copyright (c) 2018, The Holy Constituency of the Summer Triangle.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# This program is based on code from VGMToolbox.
# Copyright (c) 2009
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import struct
import io
import functools
import os
from collections import namedtuple as T
import sys

BYTE_ZERO = 0
def JOIN_BYTE_ARRAY(sr):
    return bytes(sr)

def readfunc(fmt):
    a = struct.Struct(fmt)
    b = a.size
    def f(f, at=None):
        if at is not None:
            back = f.tell()
            f.seek(at)
            d = a.unpack(f.read(b))[0]
            f.seek(back)
            return d
        else:
            return a.unpack(f.read(b))[0]

    return f

def latebinder(f):
    return lambda s: f(s.f)

class R(object):
    """ file reader based on types """
    def __init__(self, file):
        self.f = file

    int8_t    = latebinder(readfunc(">b"))
    uint8_t   = latebinder(readfunc(">B"))
    int16_t   = latebinder(readfunc(">h"))
    uint16_t  = latebinder(readfunc(">H"))
    int32_t   = latebinder(readfunc(">i"))
    uint32_t  = latebinder(readfunc(">I"))
    int64_t   = latebinder(readfunc(">q"))
    uint64_t  = latebinder(readfunc(">Q"))
    float32_t = latebinder(readfunc(">f"))

    le_int8_t    = latebinder(readfunc("<b"))
    le_uint8_t   = latebinder(readfunc("<B"))
    le_int16_t   = latebinder(readfunc("<h"))
    le_uint16_t  = latebinder(readfunc("<H"))
    le_int32_t   = latebinder(readfunc("<i"))
    le_uint32_t  = latebinder(readfunc("<I"))
    le_int64_t   = latebinder(readfunc("<q"))
    le_uint64_t  = latebinder(readfunc("<Q"))
    le_float32_t = latebinder(readfunc("<f"))

    def tell(self):
        return self.f.tell()

    def seek(self, at, where=os.SEEK_SET):
        self.f.seek(at, where)

    def struct(self, struct, at=None):
        if at is not None:
            back = self.f.tell()
            self.f.seek(at)
            d = self.struct(struct)
            self.f.seek(back)
            return d

        return struct.unpack(self.f.read(struct.size))

    def bytes(self, size, at=None):
        if at is not None:
            back = self.f.tell()
            self.f.seek(at)
            d = self.bytes(size)
            self.f.seek(back)
            return d

        return self.f.read(size)

    def bytesinto(self, inbuf, at=None):
        if at is not None:
            back = self.f.tell()
            self.f.seek(at)
            self.bytesinto(inbuf)
            self.f.seek(back)
            return

        self.f.readinto(inbuf)

    def string0(self, at=None):
        if at is not None:
            back = self.f.tell()
            self.f.seek(at)
            d = self.string0()
            self.f.seek(back)
            return d

        bk = self.f.tell()
        tl = 0
        sr = []
        while 1:
            b = self.f.read(16)
            tl += len(b)

            if len(b) == 0:
                raise Exception("EOF")

            for c in b:
                if c != BYTE_ZERO:
                    sr.append(c)
                else:
                    break
            else:
                continue
            break
        string = JOIN_BYTE_ARRAY(sr)
        self.f.seek(bk + len(string) + 1)
        return string.decode("sjis")

class Struct(struct.Struct):
    """ struct with an output filter (usually a namedtuple) """
    def __init__(self, fmt, out_type):
        struct.Struct.__init__(self, fmt)
        self.out_type = out_type

    def unpack(self, buf):
        return self.out_type(* struct.Struct.unpack(self, buf))

COLUMN_STORAGE_MASK = 0xF0
COLUMN_STORAGE_PERROW = 0x50
COLUMN_STORAGE_CONSTANT = 0x30
COLUMN_STORAGE_CONSTANT2 = 0x70
COLUMN_STORAGE_ZERO = 0x10

COLUMN_TYPE_MASK = 0x0F
COLUMN_TYPE_DATA   = 0x0B
COLUMN_TYPE_STRING = 0x0A
COLUMN_TYPE_FLOAT  = 0x08
COLUMN_TYPE_8BYTE  = 0x06
COLUMN_TYPE_4BYTE2 = 0x05
COLUMN_TYPE_4BYTE  = 0x04
COLUMN_TYPE_2BYTE2 = 0x03
COLUMN_TYPE_2BYTE  = 0x02
COLUMN_TYPE_1BYTE2 = 0x01
COLUMN_TYPE_1BYTE  = 0x00

# string and data fields require more information
def promise_data(r):
    offset = r.uint32_t()
    size = r.uint32_t()
    return lambda h: r.bytes(size, at=h.data_offset + 8 + offset)

def promise_string(r):
    offset = r.uint32_t()
    return lambda h: r.string0(at=h.string_table_offset + 8 + offset)

column_data_dtable = {
    COLUMN_TYPE_DATA   : promise_data,
    COLUMN_TYPE_STRING : promise_string,
    COLUMN_TYPE_FLOAT  : lambda r: r.float32_t(),
    COLUMN_TYPE_8BYTE  : lambda r: r.uint64_t(),
    COLUMN_TYPE_4BYTE2 : lambda r: r.int32_t(),
    COLUMN_TYPE_4BYTE  : lambda r: r.uint32_t(),
    COLUMN_TYPE_2BYTE2 : lambda r: r.int16_t(),
    COLUMN_TYPE_2BYTE  : lambda r: r.uint16_t(),
    COLUMN_TYPE_1BYTE2 : lambda r: r.int8_t(),
    COLUMN_TYPE_1BYTE  : lambda r: r.uint8_t()}

column_data_stable = {
    COLUMN_TYPE_DATA   : "8s",
    COLUMN_TYPE_STRING : "4s",
    COLUMN_TYPE_FLOAT  : "f",
    COLUMN_TYPE_8BYTE  : "Q",
    COLUMN_TYPE_4BYTE2 : "i",
    COLUMN_TYPE_4BYTE  : "I",
    COLUMN_TYPE_2BYTE2 : "h",
    COLUMN_TYPE_2BYTE  : "H",
    COLUMN_TYPE_1BYTE2 : "b",
    COLUMN_TYPE_1BYTE  : "B"}

utf_header_t = Struct(">IHHIIIHHI",
    T("utf_header_t", ("table_size", "u1", "row_offset", "string_table_offset",
    "data_offset", "table_name_offset", "number_of_fields", "row_size", "number_of_rows")))

class UTFTable(object):
    def __init__(self, file):
        buf = R(file)
        magic = buf.uint32_t()
        if magic != 0x40555446:
            raise ValueError("bad magic")

        self.header = buf.struct(utf_header_t)
        self.name = buf.string0(at=self.header.string_table_offset + 8 + self.header.table_name_offset)

        buf.seek(0x20)
        self.read_schema(buf)

        buf.seek(self.header.row_offset + 8)
        self.rows = list(self.iter_rows(buf))

    def read_schema(self, buf):
        buf.seek(0x20)

        dynamic_keys = []
        format = ">"
        constants = {}

        for _ in range(self.header.number_of_fields):
            field_type = buf.uint8_t()
            name_offset = buf.uint32_t()

            occurrence = field_type & COLUMN_STORAGE_MASK
            type_key = field_type & COLUMN_TYPE_MASK

            if occurrence in (COLUMN_STORAGE_CONSTANT, COLUMN_STORAGE_CONSTANT2):
                name = buf.string0(at=self.header.string_table_offset + 8 + name_offset)
                val = column_data_dtable[type_key](buf)
                constants[name] = val
            else:
                dynamic_keys.append(buf.string0(at=self.header.string_table_offset + 8 + name_offset))
                format += column_data_stable[type_key]

        for k in constants.keys():
            if callable(constants[k]):
                constants[k] = constants[k](self.header)

        self.dynamic_keys = dynamic_keys
        self.struct_format = format
        self.constants = constants

    def resolve(self, buf, *args):
        ret = []
        for val in args:
            if isinstance(val, bytes):
                if len(val) == 8:
                    offset, size = struct.unpack(">II", val)
                    ret.append(buf.bytes(size, at=self.header.data_offset + 8 + offset))
                else:
                    offset = struct.unpack(">I", val)[0]
                    ret.append(buf.string0(at=self.header.string_table_offset + 8 + offset))
            else:
                ret.append(val)
        return tuple(ret)

    def iter_rows(self, buf):
        sfmt = Struct(self.struct_format, functools.partial(self.resolve, buf))
        for n in range(self.header.number_of_rows):
            values = buf.struct(sfmt)
            ret = {k: v for k, v in zip(self.dynamic_keys, values)}
            ret.update(self.constants)
            yield ret

    def __repr__(self):
        return "<UTFTable '{1}' with {0} rows>".format(len(self.rows), self.name)



import sys
from setuptools import setup, Extension


def main():
    args = dict(
        ext_modules=[
            Extension(
                "_acb_speedup", sources=["fast_sub/module.c"], py_limited_api=True
            )
        ],
    )

    setup(**args)


if __name__ == "__main__":
    main()
